# Copyright 2022 Cognite AS
from indsl import (
    data_quality,
    detect,
    equipment,
    exceptions,
    filter,
    fluid_dynamics,
    forecast,
    not_listed_operations,
    oil_and_gas,
    regression,
    resample,
    signals,
    smooth,
    statistics,
    ts_utils,
)
